@@ -9,6 +9,10 @@ use tokio_stream::StreamExt;
 
 use tokio_util::codec::Framed;
 
+use async_trait::async_trait;
+
+use num::NumCast;
+
 use bytes::{Bytes, BytesMut};
 
 use blather::{codec, KVLines, Params, Telegram};
@@ -16,17 +20,20 @@ use blather::{codec, KVLines, Params, Telegram};
 use crate::err::Error;
 
 
+/// Application channel subscription identifier.
 pub enum SubCh {
   Num(u8),
   Name(String)
 }
 
+/// Channel subscription information context.
 pub struct SubInfo {
   pub ch: SubCh
 }
 
 
-/// Subscribe to an application message channel.
+/// Subscribe to an application message channel on a connection to a
+/// subscription interface.
 pub async fn subscribe<C>(
   conn: &mut Framed<C, blather::Codec>,
   subinfo: SubInfo
@@ -119,7 +126,8 @@ pub struct MsgInfo {
 }
 
 
-/// Receive a single message.
+/// Receive a single message, allowing a closure to select how to store
+/// metadata and payload based on the message header.
 ///
 /// The `storeq` is a closure that, if there's metadata and/or payload, can be
 /// called to allow the application to request how the metadata and payload
@@ -161,7 +169,7 @@ pub struct MsgInfo {
 /// // Enter an loop which keeps receiving messages until the connection is
 /// // dropped.
 /// async fn get_message(conn: &mut Framed<TcpStream, blather::Codec>) -> Msg {
-///   msg::recv(
+///   msg::recv_c(
 ///     conn,
 ///     |_mi| {
 ///       // A new message is arriving; give the application the opportunity to
@@ -175,7 +183,7 @@ pub struct MsgInfo {
 ///   ).await.unwrap()
 /// }
 /// ```
-pub async fn recv<C, S>(
+pub async fn recv_c<C, S>(
   conn: &mut Framed<C, blather::Codec>,
   storeq: S
 ) -> Result<Msg, Error>
@@ -215,7 +223,8 @@ where
 
 
 /// Enter a loop which will keep receiving messages until connection is closed
-/// or a killswitch is triggered.
+/// or a killswitch is triggered, giving the application the choice of
+/// metadata/payload storage though a closure.
 ///
 /// Returns `Ok()` if the loop was terminated by the killswitch.
 ///
@@ -237,7 +246,7 @@ where
 /// async fn get_messages(conn: &mut Framed<TcpStream, blather::Codec>) {
 ///   let mut idx = 0;
 ///
-///   msg::recvloop(
+///   msg::recvloop_c(
 ///     conn,
 ///     None,
 ///     |mi| {
@@ -277,7 +286,7 @@ where
 /// }
 /// ```
 // ToDo: yield, when it becomes available
-pub async fn recvloop<C, S, P>(
+pub async fn recvloop_c<C, S, P>(
   conn: &mut Framed<C, blather::Codec>,
   kill: Option<killswitch::Shutdown>,
   mut storeq: S,
@@ -291,7 +300,7 @@ where
   if let Some(kill) = kill {
     loop {
       tokio::select! {
-        msg = recv(conn, &mut storeq) => {
+        msg = recv_c(conn, &mut storeq) => {
           let msg = msg?;
           procmsg(msg)?;
         }
@@ -304,7 +313,7 @@ where
   } else {
     // No killswitch supplied -- just keep running until disconnection
     loop {
-      let msg = recv(conn, &mut storeq).await?;
+      let msg = recv_c(conn, &mut storeq).await?;
       procmsg(msg)?;
     }
   }
@@ -313,8 +322,8 @@ where
 }
 
 
-/// This is the same as [`recvloop()`], but it assumes the message processing
-/// closure returns a [`Future`].
+/// This is the same as [`recvloop_c()`], but it assumes the message
+/// processing closure returns a [`Future`].
 ///
 /// # Example
 ///
@@ -332,7 +341,7 @@ where
 /// async fn get_messages(conn: &mut Framed<TcpStream, blather::Codec>) {
 ///   let mut idx = 0;
 ///
-///   msg::recvloop_a(
+///   msg::recvloop_ca(
 ///     conn,
 ///     None,
 ///     |mi| {
@@ -347,7 +356,7 @@ where
 ///   ).await.unwrap();
 /// }
 /// ```
-pub async fn recvloop_a<C, S, F, P>(
+pub async fn recvloop_ca<C, S, F, P>(
   conn: &mut Framed<C, blather::Codec>,
   kill: Option<killswitch::Shutdown>,
   mut storeq: S,
@@ -362,7 +371,7 @@ where
   if let Some(kill) = kill {
     loop {
       tokio::select! {
-        msg = recv(conn, &mut storeq) => {
+        msg = recv_c(conn, &mut storeq) => {
           let msg = msg?;
           procmsg(msg).await?;
         }
@@ -375,7 +384,7 @@ where
   } else {
     // No killswitch supplied -- just keep running until disconnection
     loop {
-      let msg = recv(conn, &mut storeq).await?;
+      let msg = recv_c(conn, &mut storeq).await?;
       procmsg(msg).await?;
     }
   }
@@ -393,24 +402,7 @@ where
   C: AsyncRead + AsyncWrite + Unpin,
   S: FnMut(&MsgInfo) -> Result<(StoreType, StoreType), Error>
 {
-  let metalen = if mp.have("MetaLen") {
-    mp.get_param::<u32>("MetaLen")?
-  } else {
-    0u32
-  };
-
-  let payloadlen = if mp.have("Len") {
-    mp.get_param::<u64>("Len")?
-  } else {
-    0u64
-  };
-
-  // Parse command
-  let cmd = if mp.have("Cmd") {
-    mp.get_param::<u32>("Cmd")?
-  } else {
-    0
-  };
+  let (cmd, metalen, payloadlen) = parse_header(&mp)?;
 
   // ToDo: Parse mp and check if metadata and/or payload is passed from the
   //       server using a local file path.  If it is, then return it to the
@@ -447,70 +439,61 @@ where
   // If meta_store is Some, then request the appropriate type from the
   // blather's Codec.
   //
-  let meta = if let Some(meta_store) = meta_store {
-    match meta_store {
-      StoreType::None => {
-        // This happens if the MetaLen is non-zero, but the callback says it
-        // doesn't want the data.
-        conn.codec_mut().skip(metalen as usize)?;
-      }
-      StoreType::Bytes => {
-        conn.codec_mut().expect_bytes(metalen as usize)?;
-      }
-      StoreType::BytesMut => {
-        conn.codec_mut().expect_bytesmut(metalen as usize)?;
-      }
-      StoreType::Params => {
-        conn.codec_mut().expect_params();
-      }
-      StoreType::KVLines => {
-        conn.codec_mut().expect_kvlines();
-      }
-      StoreType::File(ref fname) => {
-        conn.codec_mut().expect_file(fname, metalen as usize)?;
-      }
-    }
+  let meta = get_content_to(conn, metalen, meta_store).await?;
 
-    get_content(conn).await?
-  } else {
-    None
-  };
-
-
-  let payload = if let Some(payload_store) = payload_store {
-    match payload_store {
-      StoreType::None => {
-        // This happens if the Len is non-zero, but the callback says it
-        // doesn't want the data.
-        conn.codec_mut().skip(payloadlen as usize)?;
-      }
-      StoreType::Bytes => {
-        conn.codec_mut().expect_bytes(payloadlen as usize)?;
-      }
-      StoreType::BytesMut => {
-        conn.codec_mut().expect_bytesmut(payloadlen as usize)?;
-      }
-      StoreType::Params => {
-        conn.codec_mut().expect_params();
-      }
-      StoreType::KVLines => {
-        conn.codec_mut().expect_kvlines();
-      }
-      StoreType::File(ref fname) => {
-        conn.codec_mut().expect_file(fname, payloadlen as usize)?;
-      }
-    }
-
-    get_content(conn).await?
-  } else {
-    None
-  };
+  let payload = get_content_to(conn, payloadlen, payload_store).await?;
 
 
   Ok(Msg { cmd, meta, payload })
 }
 
 
+/// Get content of a given size to a chosen storage type.
+///
+/// Returns `Ok(None)` if the size is zero or if the storage type is
+/// `StorageType::None`.
+async fn get_content_to<C, S>(
+  conn: &mut Framed<C, blather::Codec>,
+  size: S,
+  store_type: Option<StoreType>
+) -> Result<Option<Storage>, Error>
+where
+  C: AsyncRead + AsyncWrite + Unpin,
+  S: NumCast
+{
+  if let Some(store_type) = store_type {
+    match store_type {
+      StoreType::None => {
+        // This happens if the Len is non-zero, but the callback says it
+        // doesn't want the data.
+        conn.codec_mut().skip(num::cast(size).unwrap())?;
+      }
+      StoreType::Bytes => {
+        conn.codec_mut().expect_bytes(num::cast(size).unwrap())?;
+      }
+      StoreType::BytesMut => {
+        conn.codec_mut().expect_bytesmut(num::cast(size).unwrap())?;
+      }
+      StoreType::Params => {
+        conn.codec_mut().expect_params();
+      }
+      StoreType::KVLines => {
+        conn.codec_mut().expect_kvlines();
+      }
+      StoreType::File(ref fname) => {
+        conn
+          .codec_mut()
+          .expect_file(fname, num::cast(size).unwrap())?;
+      }
+    }
+
+    get_content(conn).await
+  } else {
+    Ok(None)
+  }
+}
+
+
 /// Translate an incoming frame from the [`blather::Codec`] into a [`Storage`]
 /// type.
 async fn get_content<C>(
@@ -535,4 +518,263 @@ where
   }
 }
 
+
+/// Implements callback methods for incoming message header and payloads.
+///
+/// The [`recv_h`] function and the [`recvloop_h`] require an object which
+/// implements this trait.
+#[async_trait]
+pub trait Handler {
+  /// Called once a message header has been received to allow the application
+  /// to choose how to store the meta and payload content.
+  async fn on_header(
+    &self,
+    mi: &MsgInfo
+  ) -> Result<(StoreType, StoreType), Error>;
+
+  /// Called once the metadata and payload have been received.
+  async fn on_data(&self, msg: Msg) -> Result<(), Error>;
+}
+
+
+/// Receive a single message, allowing a [`Handler`] method to select how to
+/// store metadata and payload based on the message header.
+///
+/// Once the message header has been received, parse it and call the supplied
+/// [`Handler`]'s [`on_header()`](Handler::on_header) method to determine how
+/// to store the message's metadata and payload content.
+pub async fn recv_h<C>(
+  conn: &mut Framed<C, blather::Codec>,
+  handler: &Box<dyn Handler + Send + Sync>
+) -> Result<Msg, Error>
+where
+  C: AsyncRead + AsyncWrite + Unpin
+{
+  // Wait for the next frame, and exiect it to be a Telegram.
+  if let Some(o) = conn.next().await {
+    let o = o?;
+    match o {
+      codec::Input::Telegram(tg) => {
+        // Got the expetected Telegram -- make sure that it's has a "Msg"
+        // topic.
+        if let Some(topic) = tg.get_topic() {
+          if topic == "Msg" {
+            // Convert to a Params buffer, since we no longer need the topic
+            let mp = tg.into_params();
+
+            return proc_inbound_msg_h(conn, mp, handler).await;
+          } else if topic == "Fail" {
+            return Err(Error::ServerError(tg.into_params()));
+          }
+        }
+      }
+      _ => {
+        return Err(Error::BadState(
+          "Unexpected codec input type.".to_string()
+        ));
+      }
+    }
+    return Err(Error::bad_state("Unexpected reply from server."));
+  }
+
+  Err(Error::Disconnected)
+}
+
+
+/// Given a `Params` buffer of a received `Msg` telegram, return a tuple
+/// containing `(command id, metadata length, payload length)`.
+fn parse_header(mp: &Params) -> Result<(u32, u32, u64), Error> {
+  let cmd = if mp.have("Cmd") {
+    mp.get_param::<u32>("Cmd")?
+  } else {
+    0
+  };
+
+  let metalen = if mp.have("MetaLen") {
+    mp.get_param::<u32>("MetaLen")?
+  } else {
+    0u32
+  };
+
+  let payloadlen = if mp.have("Len") {
+    mp.get_param::<u64>("Len")?
+  } else {
+    0u64
+  };
+
+  Ok((cmd, metalen, payloadlen))
+}
+
+
+/// Get a single message, given a parsed Msg header.
+///
+/// `mp` is the message parameters extracted from a Msg telegram.
+async fn proc_inbound_msg_h<C>(
+  conn: &mut Framed<C, blather::Codec>,
+  mp: Params,
+  handler: &Box<dyn Handler + Send + Sync>
+) -> Result<Msg, Error>
+where
+  C: AsyncRead + AsyncWrite + Unpin
+{
+  let (cmd, metalen, payloadlen) = parse_header(&mp)?;
+
+  // ToDo: Parse mp and check if metadata and/or payload is passed from the
+  //       server using a local file path.  If it is, then return it to the
+  //       application using Storage::LocalFile(PathBuf).
+
+  // If the Params contains either a Len or a MetaLen keyword, then
+  // call the application callback to determine how it wants the data
+  // stored.
+  // ToDo: - If metadata and payload are stored as "local files", then don't
+  //         call application; force to Storage::LocalFile
+  let (meta_store, payload_store) = if metalen != 0 || payloadlen != 0 {
+    // Call the application callback, passing a few Msg parameters, to ask it
+    // in what form it would like the metadata and payload.
+
+    let mi = MsgInfo {
+      cmd,
+      metalen,
+      payloadlen
+    };
+
+    let (ms, ps) = handler.on_header(&mi).await?;
+    let ms = if metalen != 0 { Some(ms) } else { None };
+    let ps = if metalen != 0 { Some(ps) } else { None };
+    (ms, ps)
+  } else {
+    (None, None)
+  };
+
+  //
+  // At this point, if meta_store is None, it means there was no message
+  // metadata, and we'll skip this altogether and return None to the
+  // application for the metadata.
+  //
+  // If meta_store is Some, then request the appropriate type from the
+  // blather's Codec.
+  //
+  let meta = get_content_to(conn, metalen, meta_store).await?;
+
+  let payload = get_content_to(conn, payloadlen, payload_store).await?;
+
+  Ok(Msg { cmd, meta, payload })
+}
+
+
+/// Enter a loop which will keep receiving messages to a [`Handler`] object
+/// until connection is closed or an optional killswitch is triggered.
+///
+/// Returns `Ok()` if the loop was terminated by the killswitch.
+///
+/// # Example
+/// Enter a loop which will receive messages:
+/// - Ignore metadata which is larder than 1MB
+/// - Ignore payloads larger than 16MB, larger than 256KB are stored in a file,
+///   and other payloads are stored in byte buffers.
+///
+/// Messages are only received, but not processed.
+///
+/// ```no_run
+/// use async_trait::async_trait;
+/// use std::path::PathBuf;
+/// use std::str::FromStr;
+/// use ddmw_client::{
+///   conn,
+///   msg::{self, recv::{StoreType, Handler, MsgInfo, Msg, SubInfo, SubCh}}
+/// };
+///
+/// struct RecvProc {
+///   idx: usize
+/// }
+///
+/// #[async_trait]
+/// impl Handler for RecvProc {
+///   async fn on_header(
+///     &self,
+///     mi: &MsgInfo
+///   ) -> Result<(StoreType, StoreType), ddmw_client::Error> {
+///     // A new message is arriving; give the application the opportunity
+///     // to choose how to store the message metadata and payload.
+///
+///     // Choose what to do with message metadata
+///     let meta_store = if mi.metalen > 1024*1024 {
+///       // Too big
+///       StoreType::None
+///     } else {
+///       // Store it in a memory buffer
+///       StoreType::Bytes
+///     };
+///
+///     // Choose what to do with message payload
+///     let payload_store = if mi.payloadlen > 16*1024*1024 {
+///       // Bigger than 16MB; too big, just ignore it
+///       StoreType::None
+///     } else if mi.payloadlen > 256*1024 {
+///       // It's bigger than 256K -- store it in a file
+///       let payloadfile = format!("{:x}.payload", self.idx);
+///       StoreType::File(PathBuf::from(payloadfile))
+///     } else {
+///       // It's small enough to store in a memory buffer
+///       StoreType::Bytes
+///     };
+///
+///     Ok((meta_store, payload_store))
+///   }
+///
+///   async fn on_data(&self, msg: Msg) -> Result<(), ddmw_client::Error> {
+///     println!("Process message");
+///
+///     // ToDo: Do things with message buffer here
+///
+///     Ok(())
+///   }
+/// }
+///
+/// async fn connect_and_receive_messages() {
+///   let pa = conn::ProtAddr::from_str("127.0.0.1:4100").unwrap();
+///   let mut conn = conn::connect(&pa, None).await.unwrap();
+///
+///   let subinfo = SubInfo { ch: SubCh::Num(11) };
+///   ddmw_client::msg::recv::subscribe(&mut conn, subinfo).await.unwrap();
+///
+///   let handler = Box::new(RecvProc { idx: 0 })
+///       as Box<dyn Handler + Sync + Send>;
+///
+///   msg::recvloop_h(&mut conn, None, &handler).await.unwrap();
+/// }
+/// ```
+// ToDo: yield, when it becomes available
+pub async fn recvloop_h<C>(
+  conn: &mut Framed<C, blather::Codec>,
+  kill: Option<killswitch::Shutdown>,
+  handler: &Box<dyn Handler + Send + Sync>
+) -> Result<(), Error>
+where
+  C: AsyncRead + AsyncWrite + Unpin
+{
+  if let Some(kill) = kill {
+    loop {
+      tokio::select! {
+        msg = recv_h(conn, handler) => {
+          let msg = msg?;
+          handler.on_data(msg).await?;
+        }
+        _ = kill.wait() => {
+          // An external termination request was received, so break out of loop
+          break;
+        }
+      }
+    }
+  } else {
+    // No killswitch supplied -- just keep running until disconnection
+    loop {
+      let msg = recv_h(conn, handler).await?;
+      handler.on_data(msg).await?;
+    }
+  }
+
+  Ok(())
+}
+
 // vim: set ft=rust et sw=2 ts=2 sts=2 cinoptions=2 tw=79 :
